/* $Id: gx.c,v 1.8 1998/11/24 18:28:24 marcus Exp $
******************************************************************************

   Cyrix MediaGX internal ramdac driver

   Copyright (C) 1998 Cyrix Corporation
   Copyright (C) 1998 Emmanuel Marty	[core@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following sources of information :

	Cyrix MediaGX confidential databooks
	Cyrix MediaGX gxvideo.c example sourcecode
	Help and code from Sengan Baring-Gould of Cyrix Corporation

	Many thanks to the Cyrix Corporation for providing me with
	documentation and official support about the GX chipset family.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define MAINTAINER	Emmanuel_Marty
#define DRIVER_NAME	"Cyrix MediaGX DAC driver"
#define DRIVER_REV	"$Revision: 1.8 $"
#define DEBUG_LEVEL	0

#include <kgi/module.h>
#include <Cyrix/gx.h>

/*	Maximum dotclock	*/

#define DAC_MAX_CLK 135000000

/*
**	RAMDAC definition for KGI
*/

static struct kgi_ramdac ramdac = {
	"Cyrix",			/* manufacturer 	*/
	"GX",				/* model		*/
	GM_TEXT16 | GM_8BIT |		/* supported modes	*/
	GM_15BIT | GM_16BIT,
	FL_DAC_GAMMA,			/* flags		*/
	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
	NULL,				/* Private data hook	*/
	DAC_16BIT_PB | DAC_8BIT_PB,	/* device type		*/
	{       0,  DAC_MAX_CLK },	/* normal mode		*/
	{       0,  0 },		/* 2xclk mode		*/

	{{	0,  DAC_MAX_CLK },	/* 	KGIGT_TEXT16	*/
	 {	0,  0 },		/* 	KGIGT_TEXT32 	*/
	 {	0,  0 },		/* 	KGIGT_1BIT		*/
	 {	0,  0 },		/* 	KGIGT_4BIT		*/
	 {	0,  DAC_MAX_CLK },	/* 	KGIGT_8BIT		*/
	 {	0,  DAC_MAX_CLK },	/* 	KGIGT_15BIT	*/
	 {	0,  DAC_MAX_CLK },	/* 	KGIGT_16BIT	*/
	 {	0, 0 },			/* 	KGIGT_24BIT	*/
	 {	0, 0 }}			/*	KGIGT_32BIT	*/
};

/*
**	set and get internal CLUT
*/

#define CLUT_MAGIC	1040		/* 2^10 + 2^4 */
#define	CLUT_SHIFT	10

void dac_set_clut (struct kgi_display *dpy, kgi_clut *clut,
		   kgi_uint first, kgi_uint cnt) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr gxbase = private->gx_control.baseptr;
	kgiu32 lockval, p;
	kgiu16 *pal = (kgiu16 *) &clut->data[first];
	
	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	WRITE_GX32 (gxbase, DC_PAL_ADDRESS, first);
	
	ASSERT (cnt);

	while (cnt--) {
		p = (((kgiu32) *pal++) >> CLUT_SHIFT) << 12;
		p |= (((kgiu32) *pal++) >> CLUT_SHIFT) << 6;
		p |= (((kgiu32) *pal++) >> CLUT_SHIFT);
		
		WRITE_GX32 (gxbase, DC_PAL_DATA, p);
	}

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

void dac_get_clut (struct kgi_display *dpy, kgi_clut *clut,
		   kgi_uint first, kgi_uint cnt) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	mem_vaddr gxbase = private->gx_control.baseptr;
	kgiu32 lockval, p;
	kgiu16 *pal = (kgiu16 *) &clut->data[first];

	lockval = READ_GX32 (gxbase, DC_UNLOCK);
	WRITE_GX32 (gxbase, DC_UNLOCK, DC_UNLOCK_VALUE);

	WRITE_GX32 (gxbase, DC_PAL_ADDRESS, first);
	
	ASSERT (cnt);

	while (cnt--) {
		p = READ_GX32 (gxbase, DC_PAL_DATA);
		*pal++ = ((p >> 12) & 0x3F) * CLUT_MAGIC;
		*pal++ = ((p >> 6)  & 0x3F) * CLUT_MAGIC;
		*pal++ = (p         & 0x3F) * CLUT_MAGIC;
	}

	WRITE_GX32 (gxbase, DC_UNLOCK, lockval);
}

/*
**	GGI interface
*/

int kgim_ramdac_init (struct kgi_card *card, struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " rev " DRIVER_REV);

	card->ramdac = &ramdac;	
	return EOK;
}

void kgim_ramdac_done (struct kgi_display *dpy) {
	NOTICE (DRIVER_NAME " removed ");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy,
			    struct kgi_mode *mode, enum kgi_tm_cmd *cmd) {
	kgi_uint gt, newclock = mode->dclk, cmul, cdiv;
	
	switch (gt = mode->request.graphtype) {
		case KGI_AUTO:
			mode->request.graphtype = gt = KGIGT_8BIT;
			break;

		case KGIGT_TEXT16:
		case KGIGT_8BIT:
		case KGIGT_15BIT:
		case KGIGT_16BIT:
			break;
			
		default:
			DEBUG1 ("graphtype %i not supported", gt);
			return -E(RAMDAC, UNKNOWN);
	}

	switch (*cmd) {
		case CMD_PROPOSE:
			mode->clk.mul = mode->lclk.mul = 1;
			mode->clk.div = mode->lclk.div = 1;
			mode->flags  &= ~FL_DAC_2XCLOCK;
			
			return EOK;
			
		case CMD_LOWER:
			if (newclock * (cmul = mode->clk.mul) /
				(cdiv = mode->clk.div) < ramdac.normal.min) {
				DEBUG2 ("DCLK too low (%i Hz)", newclock);

				return -E (RAMDAC, CLK_LOW_LIMIT);
			}
			
			if (newclock * cmul / cdiv > ramdac.normal.max)
				mode->dclk = ramdac.normal.max / cmul * cdiv;

			return EOK;

		case CMD_RAISE:
			if (newclock * (cmul = mode->clk.mul) /
				(cdiv = mode->clk.div) > ramdac.normal.max) {
				DEBUG2 ("DCLK too high (%i Hz)", newclock);

				return -E (RAMDAC, CLK_LOW_LIMIT);
			}
			
			if (newclock * cmul / cdiv < ramdac.normal.min)
				mode->dclk = ramdac.normal.min / cmul * cdiv;

			return EOK;
	
		case CMD_CHECK:
			if ((newclock * (cmul = mode->clk.mul) /
				(cdiv = mode->clk.div) < ramdac.normal.min) ||
				(newclock * cmul / cdiv > ramdac.normal.max)) {

				DEBUG2 ("%i Hz DCLK out of limits", newclock);
				return -E(RAMDAC, MODE_CHECK);
			}
		
			return EOK;
		
		default:
			INTERNAL_ERROR;
			
			return -E (RAMDAC, BOGUS);
	}
}

void kgim_ramdac_set_mode (struct kgi_display *dpy, struct kgi_mode *mode) {
	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		case KGIGT_8BIT:
			dpy->set_clut = dac_set_clut;
			break;
			
		case KGIGT_15BIT:
		case KGIGT_16BIT:
			dpy->set_clut = NULL;
			break;
			
		default:
			INTERNAL_ERROR;
			break;
	}
}

int kgim_ramdac_command (struct kgi_graphic *graph,kgi_uint cmd,kgi_uint arg) {
	kgi_clut clut;
	
	switch (cmd) {
		case RAMDAC_SETCLUT:
			clut.size = 256;
			clut.data = (kgi_color *) (graph->io_buf);
			dac_set_clut (graph->dev.dpy, &clut, 0, 256);
			return EOK;
		
		case RAMDAC_GETCLUT:
			clut.size = 256;
			clut.data = (kgi_color *) (graph->io_buf);
			graph->io_res = graph->io_buf;
			dac_get_clut (graph->dev.dpy, &clut, 0, 256);
			return EOK;
		
		default:
			return -E (RAMDAC, NOSUP_ALWAYS_CANT);	/* can't */
	}
}
			
